{                          
  Propeller Based Reverse Geo-Cache Birthday Present Project vU16
  Andrew Williams
  (WBA Consulting)
  dru@aestheticacoustics.com

SUMMARY:
    This is a propeller based Reverse Geo-Cache Project that stems from the arduino version posted on the web
    at http://arduiniana.org/projects/the-reverse-geo-cache-puzzle/ . My version uses a USB Protoboard, Parallax GPS
    module, 4D systems uOLED-128 display, Parallax standard servo, battery pack, LED, two switches, and some misc parts.
    More details on this project can be found on the thread at: http://forums.parallax.com/forums/?f=21&m=414611&g

BASIC PROGRAM FLOW CONCEPT:
    Power up and initialize
     If bypass switch is enabled, open box if it's held while showing a pixel "progress bar" on screen, then reboot after a delay
     If bypass switch is not enabled, continue normally
    Flash Happy Birthday message on screen a few times
    Show "locating" while waiting for GPS satellite lock
    Once NMEA strings begin flowing from GPS module:
     Check longitude/latitude against target location
      If not matched, display long/lat in red on screen and loop to checking location
      If matched show long/lat in green briefly, then flash "you got it" on screen before unlocking.   

TARGET LOCATION:
    38.735843,-121.252334 = Maidu Community Center in Roseville, CA (Where my wife and I had our wedding reception, 7/27/96)

DEVELOPMENT VERSIONING
    000  -  Started with Google Earth KML Logger. Modified pin assignments for my demo board setup (currently no switch, assigned to unused pin)
    001  -  Added LED2 for "is Longitude > 99" status check
    002  -  changed atoi string from 3 to 4
    003  -  Added multiple longitude routines for debugging
    004  -  Modified longitude routines for debugging
    005  -  Removed extra longitude routines, kept working one
    006  -  Resequenced longitude and latitude routines
    007  -  Resequenced longitude and latitude routines, updated header and footer in fake dat section, added switch at P0
    P08  -  Moved to PPDB, added uOLED, stripped down code
    P09  -  Modified display features
    P10  -  Added Servo functions
    P11  -  Added Matching routine to Longitude  (green value means match) and SW01 unlock bypass routine
    P12  -  Added Matching routine to Latitude, minor housekeeping, set lat/lon to Parallax, LED2 commented out
    U14  -  Moved to USB Protoboard, changed pinouts
    U15  -  This was the "5:15am on birthday morning but it works!" version. (12/21/09)
    U16  -  Final clean up/commenting for posting on completed projects forum. (1/1/10)

PROPELLER CONNECTIONS:
  P00 - LED1 Status LED, 330 ohm series resistor through LED to GND
  P02 - SW01 Bypass Switch, 10k pullup, switch pulls to GND
  P04 - Servo data pin
  P05 - uOLED Reset Pin
  P06 - uOLED TX (connected to uOLED module RX pin and 10k pullup)
  P07 - uOLED RX (connected to uOLED module TX Pin)
  P16 - GPS Module Output pin (connected to GPS Module SIO pin)
  P17 - Not Connected

OTHER CONNECTIONS:
  uOLED Vcc - +3.3V
  GPS VCC   - 5V
  Servo VCC - 5V
  ALL GNDs  - Vss
  Battery pack routes through main power switch then to protoboard input
  

GPS portions started from Google Earth KML Logger by "paul_h"

}
        
CON

  _CLKMODE            = XTAL1 + PLL16X
  _XINFREQ            = 5_000_000
  _stack              = 48

' GPS Assignments                                        
  GPSin     = 16                                        ' Data string from Parallax GPS module. GPS module has RAW pin pulled low
  GPSout    = 17                                        ' Only used by GPS object, not connected to anything
  SW01      = 02                                        ' bypass switch, 10k pullup on I/O pin

  LED1      = 00                                        ' status LED (mainly for debugging)
                                                                                                                
VAR
  byte s,d,datum[64]                                    ' space for file copying 
  long SAddr
  long MAddr
  word Type
  word HW
  word SW
  word DevX
  word DevY
  word red
  word green
  word lonok
  word latok
    
OBJ                                                    
  gps        :  "GPS_IO_mini_PNAV"                      ' Services GPS input           
  dm         :  "DynamicMathLib"                        ' MK's cog-reduced code
  fmt        :  "format"                                ' Used for reformatting data between NUM, STR, types.
  num        :  "Numbers"
  OLED       :  "uOLED-128-GMD1"                        ' uOLED 128 display by Steve McManus, Pins set to RE=5, TX=6, and RX=7 in this file
  DELAY      :  "Clock"
  servo      :  "servoEngine1"                          ' Servo Engine by Kye. Servo pin set to 4 in this file
 
PUB start           ' get everything going

''Initialize
dira[LED1]~~                                            ' set LED to be output
dira[SW01]~                                             ' set switch to be input

  lonok := 0                                            ' set Longitude Match flag to 0 (false)
  latok := 0                                            ' set Latitude Match flag to 0 (false)
  
  servo.servoEngine                                     ' start servo engine in it's own cog
  servo.setPositionA(180)                               ' ensure servo is in full lock position (power-up jitter may have nudged it)
  gps.start(GPSin,GPSout)                               ' start GPS streaming data update in its own cog
  OLED.INIT                                             ' wake up uOLED display
  DELAY.PauseSec(1)                                     ' take a break to let things settle (had random display issues without this pause)
  SETUP                                                 ' setup the display
  
   if INA[SW01] == 1                                    ' check bypass switch status
     happybday                                          ' if still pulled high, show happy birthday message
   else
     progressbar                                        ' if pulled low by bypass switch, go to progressbar routine
 
  OLED.FTEXT(2,3, 2, 100,255,255, string("locating...."),0)   ' display locating on screen while waiting for GPS lock
  DELAY.PauseSec(1)
  CheckForSatellite                                     ' make sure GPS is active before continuing
  waitcnt(12_000_000 + cnt)                             ' stabilize, housekeeping loop cycle time (~1 sec)

  SETUP                                                 ' setup the display for new data
num.init                                                ' start numbers and math routines
ControlLoop                                             ' main program loop

PUB ControlLoop                  ' main program loop
 checklatlon                     ' check/display current position
 DELAY.PauseSec(1)               ' loop delay, 
 ControlLoop

 
   
PUB SETUP                        ' setup the display for new data  (blackout the display)
  OLED.ERASE                     ' duh
  DELAY.PauseMSec(20)            ' give it some time
  OLED.BACKGROUND(0,0,0)         ' set background to black (RGB = 0,0,0)

PUB HAPPYBDAY                    ' Displays Happy Birthday message on uOLED display
  
    OLED.INIT                       
    DELAY.PauseSec(1)
    SETUP
    DELAY.PausemSec(300)
  repeat 10                                                ' "blink" happy birthday text on screen 
    OLED.FTEXT(2,3, 2, 100,255,255, string("Happy"),0)     ' happy @ column 2, row 3, aqua
    OLED.FTEXT(4,5, 2, 255,100,255, string("Birthday"),0)  ' birthday @ column 4, row 5, pink
    OLED.FTEXT(6,7, 2, 255,255,100, string("Chelle!"),0)   ' Chelle @ column 6, row 7, yellow   (Chelle is short for Michelle)
    DELAY.PausemSec(300)                                   ' wait a bit with text on screen
    oled.erase                                             ' erase text
    DELAY.PausemSec(300)                                   ' wait a bit with a blank screen
  DELAY.PauseSec(2)                                          
  
  SETUP                                                    ' clear screen before continuing

PUB checklatlon | idx, DD, d8, decDD, temp, match, i       ' routine for checking location against target position and displaying it on screen 
                                                        
 OLED.erase                                                ' clear screen
 OLED.FTEXT(2,1, 2, 50,255,50, string("Current"),0)        ' display title, 1st row
 OLED.FTEXT(2,2, 2, 50,255,50, string("Location:"),0)      ' display title, 2nd row

'Longitude Matching
' OLED.FTEXT(0,3, 2, 255,0,20, string(">-121.252334"),0)   ' show target longitude on screen (used during debugging)

  d8 := atoi(gps.longitude+3,2) * 100_000 / 6           ' 8 digit, frac degrees. intermediate needed before conv to string for accuracy.
  d8 += atoi(gps.longitude+6,4) * 10 /6                 ' 
  decDD    := num.ToStr(d8,Num#DEC7)                    ' string of the decimal component

   If atoi(gps.longitude,3) == 121 AND ||(252334 - d8) < 100     ' is the longitude 121 for whole number and within .0001 for decimal?
      red := 0                                          ' if it is, set red variable to 0, green to 255, and longitude ok flag to true (1)
      green := 255
      lonok := 1
   Else
      red := 255                                        ' if not, set red variable to 255, green to 0, and longitude ok flag to false (0)
      green := 0
      lonok := 0

'Longitude Display        
  DD     := num.ToStr(atoi(gps.longitude,3),Num#DEC4)   ' whole degrees - DEC# is the number of digits needed in the string
  OLED.FTEXT(1,4, 2, red,green,0, string("-"),0)        ' Correct sign for western hemisphere, manually for now
   if atoi(gps.longitude,3) > 99                        ' if longitude has 3 digits......
     OLED.FTEXT(1,4, 2, red,green,0, num.ToStr(atoi(gps.longitude,3),Num#DEC4),0)  ' display 3 digit whole degrees value in red or green
   else
     OLED.FTEXT(2,4, 2, red,green,0, num.ToStr(atoi(gps.longitude,3),Num#DEC3),0)  ' display 2 digit whole degrees value (this line not tested) 
  OLED.FTEXT(5,4, 2, red,green,0, string("."),0)        ' manually add decimal point
  d8 := atoi(gps.longitude+3,2) * 100_000 / 6           ' 8 digit, frac degrees. intermediate needed before conv to string for accuracy.
  d8 += atoi(gps.longitude+6,4) * 10 /6                 ' 
  decDD    := num.ToStr(d8,Num#DEC7)                    ' string of the decimal component
  OLED.FTEXT(5,4, 2, red,green,0,decDD,0)               ' display decimal portion of longitude in red or green

'Latitude Matching                                       
' OLED.FTEXT(0,6, 2, 255,0,20, string("> 38.735843"),0)  ' show target latitude on screen (used during debugging)

  d8 := atoi(gps.latitude+2,2) * 100_000 / 6            ' intermediate needed before conv to string for accuracy.
  d8 += atoi(gps.latitude+5,4) * 10 /6             
  decDD    := num.ToStr(d8,Num#DEC7)                    ' 7byte string of the decimal component

   If atoi(gps.latitude,2) == 38 AND ||(735843 - d8) < 100      ' is the latitude 38 for whole number and within .0001 for decimal? 
      red := 0                                          ' if it is, set red variable to 0, green to 255, and latitude ok flag to true (1)
      green := 255
      latok := 1
   Else
      red := 255                                        ' if not, set red variable to 255, green to 0, and latitude ok flag to false (0)
      green := 0
      latok := 0
 
'Latitude Display
  DD     := num.ToStr(atoi(gps.latitude,2),Num#DEC3)    ' whole degrees - write as is. It is only 2 digits
  OLED.FTEXT(1,7, 2, red,green,0, num.ToStr(atoi(gps.latitude,2),Num#DEC3),0) ' display 2 digit whole degrees value in red or green
  OLED.FTEXT(4,7, 2, red,green,0, string("."),0)        ' manually add decimal point
  d8 := atoi(gps.latitude+2,2) * 100_000 / 6            ' intermediate needed before conv to string for accuracy.
  d8 += atoi(gps.latitude+5,4) * 10 /6             
  decDD    := num.ToStr(d8,Num#DEC7)                    ' 7byte string of the decimal component
  OLED.FTEXT(4,7, 2, red,green,0,decDD,0)               ' display decimal portion of latitude in red or green

  if latok == 1 AND lonok == 1                          ' if longitude and latitude flags are true.....
    surprise                                            ' go to surprise routine to "claim prize"

                                                        ' returns to control loop
    
PUB Progressbar | Temp                    ' shows dotted prgress bar while bypass switch is held at bootup
   
  REPEAT Temp from 10 to 120 step 2       ' step from position 10 to 120 in the x axis of screen (left to right)
   OLED.PUT_PIXEL(Temp,110, 255,0,0)      ' put a red pixel at the position
    IF INA[SW01] == 1                     ' is bypass switch no longer enabled?
      reboot                              ' then bypass is aborted, so reboot
   DELAY.PauseMSec(100)                   ' pause before next pixel

  if INA[SW01] == 0                       ' if bypass is still triggered at the end of the progressbar
    unlock                                ' unlock the box
  else                                    ' this section can probably be removed since it shouldn't be needed 
    OLED.erase                            ' clear screen
    OLED.FTEXT(0,0,2, 0,255,0, string("error"),0)   ' show error on screen since bypass was not followed through
    DELAY.PauseSec(4)                     ' leave "error" on screen shortly
  reboot                                  ' just to be safe, kill this path with a fresh boot (had some issues during debugging)
   
Pub Surprise                              ' shows a message before opening when the box is in the target location
             
    OLED.INIT                             ' reset comms to display
    DELAY.PauseSec(1)
    SETUP                                 ' blackout display
    DELAY.PausemSec(300)
    
  repeat 25                               ' blink "winning" message 25 times
    oled.erase
    DELAY.PausemSec(200)
    OLED.FTEXT(2,3, 2, 100,255,255, string("YOU"),0)       ' you @ column 2, row 3, aqua
    OLED.FTEXT(4,5, 2, 255,100,255, string("GOT"),0)       ' got @ column 4, row 5, pink
    OLED.FTEXT(6,7, 2, 255,255,100, string("IT!"),0)       ' it! @ column 6, row 7, yellow
    DELAY.PausemSec(200)
  DELAY.PauseSec(5)                                       ' leave the message up for 5 seconds before opening   
  unlock                                                  ' unlock the box

  SETUP


Pub unlock                                ' unlocks the box 

    OLED.erase                            
    OLED.FTEXT(3,3, 2, 0,255,0, string("OPENING"),0)    ' display open message
    servo.setPositionA(0)                               ' move servo to 0 position to unlock
    DELAY.PauseSec(10)                                  ' leave open for 10 seconds 
    reboot                                              ' reboot (will re-lock during inital boot)
  
Pub CheckForSatellite | valid, i              ' this section is clean
repeat
   toggle(LED1)                               ' blink LED while waiting

while byte[gps.valid] <> "A"                  ' look for correct data stream from GPS indicating GPS lock (>3 satellites found) 


PUB atoi( pptr,c)| ptrr, sign                           ' convert ascii characters into number 
  result := sign := 0
  if byte[pptr] == "-"
      sign++
      pptr++
  c--    
  repeat ptrr from 0 to c
    if byte[pptr+ptrr] == 0                             ' stop if null
         quit
         
    if byte[pptr+ptrr] == "."                           ' stop at decimal point
         quit
    else     
       result := result * 10 + (byte[pptr+ptrr] - "0")
  if sign == 1
    result := -result
          

PUB Toggle(pin)                   ' toggle output pin (used to blink LED)
  dira[Pin]~~
  outa[pin]~~                                                            
  repeat 2
    !outa[Pin]
    waitcnt(500_0000 + cnt)